window {
    background-color: rgba (36, 47, 79, 0.92);
    color: #eeeeee
}

list {
    background: none;
    border-radius: 15px
}

entry {
    background-color: rgba (0, 0, 0, 0.2)
}

button {
    background: none;
    border: none
}

button:hover {
    background-color: rgba (255, 255, 255, 0.1)
}